from functools import partial

import krippendorff

from cml.domain.reconstruction import PragmaticMachineLearningModel
from cml.domain.data_source import DataSource
from cml.domain.knowledge import KnowledgeDatabase, RelativeFinder, TS_QUEUE
from cml.shared.settings import DeconstructionSettings
from cml.shared.errors import (
    DeconstructionFailed,
    NoModelReconstructedError,
    NotEnoughFeaturesWarning)


__all__ = (
    "Deconstructor",
)


def log(func):
    def wrapper(self, tier, pragmatic, learnblock):
        ###################################################################
        self.logger.protocol("{:*^100}".format("Deconstruction"))
        ###################################################################
        return func(self, tier, pragmatic, learnblock)
    return wrapper


class Deconstructor:
    TIME_COLUMN = "T"
    SUBJECT_COLUMN = "Sigma"
    PURPOSE_COLUMN = "Z"

    def __init__(self,
                 mode,
                 knowledge_database: KnowledgeDatabase,
                 relative_finder: RelativeFinder,
                 source: DataSource,
                 settings: DeconstructionSettings):
        self.mode = mode
        self.source = source
        self.settings = settings
        self.relative_finder = relative_finder
        self.knowledge_database = knowledge_database
        self.next_model_counter = knowledge_database.model_counter()
        self.logger = None
        self.reconstructor = None

    def _strategies(self, block):
        # yield (("complete", ), partial(self.deconstruct_complete, block=block))
        yield (("Sigma", "Z"), partial(self.deconstruct_sigma_zeta, block=block))
        yield (("T", "Sigma"), partial(self.deconstruct_time_sigma, block=block))
        # yield (("T", "Z"), partial(self.deconstruct_time_zeta, block=block))

    @log
    def deconstruct(self,
                    tier: int,
                    prag_model: PragmaticMachineLearningModel,
                    learnblock) -> None:

        success = False
        ###################################################################
        self.logger.protocol(self.settings)
        self.logger.protocol("{:<20}: {}".format("Pragmatic", str(prag_model)))
        ###################################################################
        for pair, strategy in self._strategies(learnblock):

            ###################################################################
            self.logger.protocol("{:<20}".format(str(pair)))
            ###################################################################

            for relative in self.relative_finder.find(pair, prag_model):

                ###############################################################
                self.logger.protocol(
                    "{:<20}: {}".format("Relative", str(relative)))
                ###############################################################

                try:
                    strategy(tier, prag_model, relative)
                    success = True
                    if self.settings.deconst_mode == "minimal": return

                except NoModelReconstructedError:
                    continue

                except DeconstructionFailed:
                    continue

                except Exception as error:
                    print(error.with_traceback())

        if not success:
            # All deconstructions failed, so save the pragmatic model
            self.knowledge_database.insert(prag_model)

    def deconstruct_time_sigma(self,
                               tier: int,
                               p_model: PragmaticMachineLearningModel,
                               r_model: PragmaticMachineLearningModel,
                               block):
        success = False
        if r_model and p_model.tier < self.settings.highest_tier-1:
            second_block = r_model.trained_with(self.source)
            # overlapping_times = set(second_block.get_column_values("T")).intersection(
             #   set(block.get_column_values("T")))
            #overlapp_a = block.new_block_from(overlapping_times)
            #overlapp_b = second_block.new_block_from(overlapping_times)
            # biggest = max(overlapp_a, overlapp_b)
            # overlapping = second_block.new_block_from(block.get_column_values("T"))

            overlapping = block.get_overlapping(second_block)
            if len(overlapping) >= self.settings.learn_block_minimum:
                # TODO (dmt): What to do if the pre images have different size?
                #overlapping_b = block.new_block_from(overlapping.get_column_values("T"))

                alpha = self.calculate_reliability(
                    overlapping.get_column_values("Z"),
                    overlapping.get_column_values("Z_other"))
                alpha_systematic = alpha < 0
                alpha_weak_reliability = 0 <= alpha < self.settings.min_reliability
                if (self.settings.allow_weak_reliability and
                   alpha_weak_reliability) or alpha_systematic:
                    # overlapping_b = block.new_block_from(overlapping.get_column_values("T"))
                    overblock = self.source.new_learnblock(
                        values=list(zip(
                            overlapping.get_column_values("Z"),
                            overlapping.get_column_values("Z_other"),
                            overlapping.get_column_values("T"),
                            ("\"\"" for _ in range(overlapping.rows)),
                            ("\"\"" for _ in range(overlapping.rows)))),
                        columns=(p_model.uid, r_model.uid, "T", "Sigma", "Z"),
                        index=[i for i in range(overlapping.rows)],
                        origin=[p_model.uid, r_model.uid]
                    )

                    # samples
                    # data = list(zip(over_block.get_column_values("Z"),
                    #                 over_block.get_column_values("T"),
                    #                 ["\"\"" for _ in range(over_block.rows)],
                    #                 ["\"\"" for _ in range(over_block.rows)]))
                    # feature = ".".join(["0", str(tier+1), "1"])
                    # columns = [feature, "T", "Sigma", "Z"]
                    # source = self.source.new_learnblock(
                    #     values=data, columns=columns, index=over_block.indexes,
                    #     origin=[p_model.uid, r_model.uid])
                    TS_QUEUE.append((tier+1, overblock))
                    success = True

        if not success:
            raise DeconstructionFailed()

    def deconstruct_time_zeta(self,
                              tier: int,
                              prag_model: PragmaticMachineLearningModel,
                              relative_model: PragmaticMachineLearningModel,
                              block):
        success = False
        self.knowledge_database.insert(prag_model)

        if relative_model:
            # Get learnblock that trained relative model
            second_block = relative_model.trained_with(self.source)

            # Get samples that have overlapping timestamp
            over_block = block.overlapping_rows(second_block, subset=["T"])
            if over_block.rows >= self.settings.learn_block_minimum:

                # Create new metadata for a pragmatic model
                new_model = prag_model.fusion(
                    relative_model, self.next_model_counter(tier))

                # Which models should be used for the reconstruction
                which_ml_models = new_model.sigma

                # Get learningblock
                train_block = block.fusion(second_block)

                # Start the reconstruction
                try:
                    recon_model = self.reconstructor.reconstruct(
                        tier, train_block, which_ml_models, new_model)
                    self.knowledge_database.replace(relative_model, recon_model)
                    success = True
                except (NoModelReconstructedError, NotEnoughFeaturesWarning):
                    pass
                except ValueError as error:
                    print(error.with_traceback())

                # alpha = self.calc_reliability(relative_model, prag_model, block)
                # if alpha >= self.settings.min_reliability:

        if not success:
            raise DeconstructionFailed()

    def deconstruct_sigma_zeta(self,
                               tier: int,
                               p_model: PragmaticMachineLearningModel,
                               r_model: PragmaticMachineLearningModel,
                               block):
        success = False

        if r_model and self.time_constraint(p_model, r_model, "SigmaZ"):

            # Get learnblock that trained relative model
            second_block = r_model.trained_with(self.source)

            overlapping_block = block.same_features_fusion(second_block)

            # Check constraint
            if overlapping_block.n_features >= 2:
                # Model fusion
                new_model = p_model.fusion(r_model, tier, self.next_model_counter(tier))
                which_ml_models = new_model.sigma
                try:
                    # Reconstruct model
                    recon_m = self.reconstructor.reconstruct(
                        tier, overlapping_block, which_ml_models, new_model)

                    self.knowledge_database.replace(r_model, recon_m)
                    success = True

                except (NoModelReconstructedError, NotEnoughFeaturesWarning):
                    if self.settings.deconst_mode == "conservative":
                        self.knowledge_database.remove(p_model)
                    elif self.settings.deconst_mode == "integrative":
                        self.knowledge_database.insert(p_model)
                    elif self.settings.deconst_mode == "oppurtunistic":
                        if block.rows > second_block.rows:
                            self.knowledge_database.remove(r_model)
                            self.knowledge_database.insert(p_model)
                        else:
                            self.knowledge_database.remove(p_model)
                            self.knowledge_database.insert(r_model)

        if not success:
            raise DeconstructionFailed()

    def time_constraint(self,
                        prag_model: PragmaticMachineLearningModel,
                        relative_model: PragmaticMachineLearningModel,
                        _type: str):

        if _type == "SigmaZ":
            if self.settings.deconst_max_distance_t == 0:
                return self._overlapping_time(prag_model, relative_model)

            elif 0 < self.settings.deconst_max_distance_t < 1:
                return self._enclosing_time(prag_model, relative_model)

            elif self.settings.deconst_max_distance_t == 1:
                return True
        elif _type == "complete":
            return prag_model.min_timestamp >= relative_model.max_timestamp \
                   >= prag_model.max_timestamp \
                   or prag_model.min_timestamp <= relative_model.min_timestamp \
                   and prag_model.max_timestamp >= relative_model.max_timestamp

    def _overlapping_time(self,
                          prag_model: PragmaticMachineLearningModel,
                          relative_model: PragmaticMachineLearningModel):

        return relative_model.min_timestamp >= prag_model.max_timestamp or \
            prag_model.min_timestamp >= relative_model.max_timestamp

    def _enclosing_time(self, prag_model, relative_model):
        m_dash_max_time = max(prag_model.max_timestamp,
                              relative_model.max_timestamp)
        m_dash_min_time = min(prag_model.min_timestamp,
                              relative_model.min_timestamp)
        relative_m_dash_time = self.settings.deconst_max_distance_t*(
                m_dash_max_time-m_dash_min_time)

        return (relative_model.min_timestamp - prag_model.max_timestamp) < \
            relative_m_dash_time or (prag_model.min_timestamp -
            relative_model.max_timestamp) < relative_m_dash_time

    def deconstruct_complete(self,
                             tier: int,
                             p_model: PragmaticMachineLearningModel,
                             r_model: PragmaticMachineLearningModel,
                             block):
        success = False

        try:
            # Check feature intersection constraint
            if r_model and self._feature_intersection(p_model, r_model) >= 2:
                new_model = p_model.fusion(
                    r_model, self.next_model_counter(tier))

            # Check time contraint
            elif r_model and self.time_constraint(p_model, r_model, "complete"):

                # target match?

                # Create submodel from TSgima relative samples
                second_block = r_model.trained_with(self.source)
                new_block = block.same_features_fusion(second_block)
                ts_relatives = self.source.time_sigma_relatives(new_block)
                which_ml_models = p_model.subject + r_model.subject
                self.reconstructor.reconstruct(
                    tier, ts_relatives, which_ml_models)
                new_model = p_model.fusion(r_model, self.next_model_counter(tier))
            else: return

            # Create learnblock
            first_block = p_model.trained_with(self.source)
            second_block = r_model.trained_with(self.source)
            new_block = first_block.same_features_fusion(second_block)
            which_ml_models = new_model.sigma

            try:
                # Reconstruct model
                recon_model = self.reconstructor.reconstruct(
                    tier, new_block, which_ml_models, new_model)
                self.knowledge_database.remove(r_model)
                self.knowledge_database.insert(recon_model)
                success = True

            except (NoModelReconstructedError, NotEnoughFeaturesWarning):
                success = self.model_differentiation(tier, new_block, r_model)

        except (NoModelReconstructedError, NotEnoughFeaturesWarning):
            self.knowledge_database.remove(r_model)
            self.knowledge_database.insert(p_model)

        finally:
            if not success:
                raise DeconstructionFailed()

    def model_differentiation(self,
                              tier: int,
                              block,
                              relative_model: PragmaticMachineLearningModel):
        success = False

        time_column = block.get_column_values("T")
        density = self.source.estimate_density(time_column)
        self.source.remove_time_dense_relatives(block, density)
        clusters = self.source.cluster(block, density)
        for time_values in clusters:
            learnblock = block.new_block_from(time_values)
            try:
                reconstructed_model = self.reconstructor.reconstruct(
                    tier, learnblock)
                self.knowledge_database.insert(reconstructed_model)
                success = True
            except NoModelReconstructedError:
                self.knowledge_database.remove_dependent_models(relative_model)

        if not success:
            self.knowledge_database.remove(relative_model)

        return success

    def _feature_intersection(self,
                              prag_model: PragmaticMachineLearningModel,
                              relative_model: PragmaticMachineLearningModel):
        first_block = prag_model.trained_with(self.source)
        second_block = relative_model.trained_with(self.source)
        return len(
            set(first_block.columns()).intersection(set(second_block.columns()))
        )

    def calculate_reliability(self, predicts_a, predicts_b):
        # TODO (dmt): Krippendorff.alpha needs numpy float64 values to work
        # correct or an TypeError exception is thrown.
        predicts_b = [int(i) for i in predicts_b]
        predicts_a = [int(i) for i in predicts_a]
        predictions = [predicts_b, predicts_a]
        if self.reconstructor.mode == "conceptual":
            try:
                return krippendorff.alpha(predictions, level_of_measurement="nominal")
            except TypeError:
                print(len(predicts_a))
                print(len(predicts_b))
                print(predictions)
                input()
                raise Exception()
        elif self.reconstructor.mode:
            return krippendorff.alpha(predictions, level_of_measurement="ration")
    #
    # def calc_reliability(self,
    #                      model_a: PragmaticMachineLearningModel,
    #                      model_b: PragmaticMachineLearningModel,
    #                      block):
    #     y_one = model_a.model.predict(block.as_numpy_array())
    #     y_two = model_b.model.predict(block.as_numpy_array())
    #     reliability_data = [y_one, y_two]
    #     if self.reconstructor.category == "conceptual":
    #         return krippendorff.alpha(reliability_data,
    #                                   level_of_measurement="nominal")
    #     elif self.reconstructor.category == "procedural":
    #         return krippendorff.alpha(reliability_data,
    #                                   level_of_measurement="ratio")
    #     else:
    #         raise ValueError()


