import pickle
from collections import deque
from abc import ABC
from collections import defaultdict
from typing import Tuple, Optional

from cml.domain.reconstruction import PragmaticMachineLearningModel


TS_QUEUE = deque()


def notify_inverted_index(func):
    def wrapper(self, *args, **kwargs):
        for obs in self.observer:
            getattr(obs, func.__name__)(*args, **kwargs)
        return func(self, *args, **kwargs)
    return wrapper


def notify_deque(func):
    def wrapper(self, *args, **kwargs):
        # replace
        if len(args) == 2:
            model, *_ = args
        # remove
        else: model, = args
        deque_copy = TS_QUEUE.copy()
        for tier, learnblock in deque_copy:
            if model.uid in learnblock.origin:
                TS_QUEUE.remove((tier, learnblock))
        return func(self, *args, **kwargs)
    return wrapper


class KnowledgeDatabase(ABC):
    def __init__(self, highest_tier):
        self.database = [KnowledgeDomain(i) for i in range(highest_tier)]
        self.observer = []
        self.n_models = 0
        super().__init__()

    def __contains__(self, item: PragmaticMachineLearningModel):
        if not isinstance(item, PragmaticMachineLearningModel):
            raise TypeError()
        try:
            self.get(item.uid)
            return True
        except KeyError:
            return False

    @notify_inverted_index
    def insert(self, model: PragmaticMachineLearningModel):
        if model not in self:
            self.database[model.tier].insert(model)
            self.n_models += 1

            ###################################################################
            self.logger.protocol("{:<20}: {}".format("Inserted", model))
            ###################################################################

    @notify_deque
    @notify_inverted_index
    def remove(self, model: PragmaticMachineLearningModel):
        self.database[model.tier].remove(model)
        self.n_models -= 1

        ###################################################################
        self.logger.protocol("{:<20}: {}".format("Removed", model))
        ###################################################################

    @notify_deque
    @notify_inverted_index
    def replace(self,
                replaced: PragmaticMachineLearningModel,
                replacer: PragmaticMachineLearningModel):
        self.database[replaced.tier].replace(replaced, replacer)
        ###################################################################
        self.logger.protocol("{:<20}: {} {:<20}: {}".format(
            "Replaced", str(replaced), "with", str(replacer)))
        ###################################################################

    @notify_inverted_index
    def extend(self): pass

    def get(self, uid: str):
        _, tier, _ = uid.split(".")
        return self.database[int(tier)].get(uid)

    def dump(self, path: str):
        file = open(path, "wb")
        pickle.dump(self, file)
        file.close()

    def model_counter(self):
        def counts(tier):
            return self.database[tier].biggest_id + 1
        return counts

    def remove_dependent_models(self,
                                relative_model: PragmaticMachineLearningModel):
        for domain in self.database:
            for model in domain.knowledge.values():
                if model.origin == relative_model.origin:
                    self.remove(model)

    def inject_logger(self, logger):
        self.logger = logger


class ConceptualKnowledgeDatabase(KnowledgeDatabase):
    def __init__(self, highest_tier):
        super().__init__(highest_tier)


class ProceduralKnowledgeDatabase(KnowledgeDatabase):
    def __init__(self, highest_tier):
        super().__init__(highest_tier)


class KnowledgeDomain:
    def __init__(self, tier: int):
        self.tier = tier
        self.knowledge = {}
        self.biggest_id = 0

    def get(self, uid: str):
        return self.knowledge[uid]

    def insert(self, model: PragmaticMachineLearningModel):
        self.knowledge[model] = model
        self.update_biggest_id(model.counter)

    def remove(self, model: PragmaticMachineLearningModel):
        del self.knowledge[model]

    def replace(self,
                replaced: PragmaticMachineLearningModel,
                replacer: PragmaticMachineLearningModel):
        del self.knowledge[replaced]
        self.knowledge[replacer] = replacer
        self.update_biggest_id(replacer.counter)

    def update_biggest_id(self, new_id):
        self.biggest_id = max(self.biggest_id, new_id)


class RelativeFinder:
    def __init__(self):
        self.index_t = defaultdict(set)
        self.index_z = defaultdict(set)
        self.index_sigma = defaultdict(set)

    def find(self,
             pair: Tuple[str, Optional[str]],
             model: PragmaticMachineLearningModel):

        if pair == ("T", "Z"):
            set_t = self.index_t[(model.min_timestamp, model.max_timestamp)]
            set_z = self.index_z[model.purpose]
            relatives = set_t.intersection(set_z)
        elif pair == ("T", "Sigma"):
            set_t = self.index_t[(model.min_timestamp, model.max_timestamp)]
            set_s = self.index_sigma[model.subject]
            relatives = set_t.intersection(set_s)
        elif pair == ("Sigma", "Z"):
            set_s = self.index_sigma[model.subject]
            set_z = self.index_z[model.purpose]
            relatives = set_s.intersection(set_z)
        elif pair == ("complete", ):
            set_t = self.index_t[(model.min_timestamp, model.max_timestamp)]
            set_s = self.index_sigma[model.subject]
            set_z = self.index_z[model.purpose]
            relatives = set_t.intersection(set_s).intersection(set_z)
        else: raise ValueError

        if model in relatives:
            relatives.remove(model)
        for relative in relatives:
            yield relative
        if not relatives:
            yield None

    def remove(self, model: PragmaticMachineLearningModel):
        t_list, s_list, z_list = self.get_index_lists(model,
                                                      time=True,
                                                      subject=True,
                                                      purpose=True)
        t_list.remove(model)
        s_list.remove(model)
        z_list.remove(model)

    def replace(self,
                replaced: PragmaticMachineLearningModel,
                replacer: PragmaticMachineLearningModel):
        t_list, s_list, z_list = self.get_index_lists(replaced,
                                                      time=True,
                                                      subject=True,
                                                      purpose=True)
        t_list.remove(replaced)
        s_list.remove(replaced)
        z_list.remove(replaced)
        self.index_t[(replacer.min_timestamp, replacer.max_timestamp)].add(
            replacer)
        self.index_z[replacer.purpose].add(replacer)
        self.index_sigma[replacer.subject].add(replacer)

    def insert(self, model: PragmaticMachineLearningModel):
        self.index_t[(model.min_timestamp, model.max_timestamp)].add(model)
        self.index_z[model.purpose].add(model)
        self.index_sigma[model.subject].add(model)

    def get_index_lists(self,
                        model: PragmaticMachineLearningModel,
                        time: bool,
                        subject: bool,
                        purpose: bool):
        t_list = s_list = z_list = None
        if time:
            t_list = self.index_t[(model.min_timestamp, model.max_timestamp)]
        if subject:
            s_list = self.index_sigma[model.subject]
        if purpose:
            z_list = self.index_z[model.purpose]
        return t_list, s_list, z_list

    def extend(self): pass


